#include <stdlib.h>
#include <stdio.h>
#include <vector>
#include <list>

using std::vector;
using std::list;

#define MAX_LINE_LEN 1024

int CountLinesInFile(char *filename)
{
    FILE *fIN=NULL;
    char dummy[MAX_LINE_LEN];
    int count=0;
    
    if ((fIN=fopen(filename,"r")) == 0) {
        fprintf(stderr,"CountLinesInFile: can't open %s\n",filename);
        exit(1);
    }
    while (fgets(dummy,MAX_LINE_LEN,fIN)!=NULL) count++;

    fclose(fIN);
    return count;
}   
    
int CountWordsInFile(char *filename)
{
    FILE *fIN;
    char dummy[MAX_LINE_LEN];
    
    if ((fIN=fopen(filename,"r")) == 0) {
        fprintf(stderr,"CountWordsInFile: can't open %s\n",filename);
        return -1;
    }
    int count=1;
    if (fscanf(fIN,"%s",dummy) == EOF) {
        fprintf(stderr,"CountWordsInFile: %s is empty!\n",filename);
        return -1;
    }
    while(fscanf(fIN,"%s",dummy)!=EOF)
        count++;
    
    fclose(fIN);
    return count;
}    

void AddEdge(vector<list<int> > &Edges, int val1, int val2) {
    int a, b;
    if ( val1 < val2 ) {
        a = val1;
        b = val2;
    } else {
        b = val1;
        a = val2;
    }
    Edges[a].push_back(b);
}

int main(int argc, char *argv[]) {

    // Check command line arguments.
    //
    if ( argc !=2 ) {
        fprintf(stderr,"Need ONE-OFFSET input faces text file.\n");
        exit(1);
    }

    // Count number of words and lines.
    // The number of lines "should" be the number of faces.
    //
    int numWords = CountWordsInFile(argv[1]);
    int numLines = CountLinesInFile(argv[1]);
    int numFaces = numLines;
    printf("Number of faces    = %d\n",numFaces);

    // Allocate memory for Faces matrix.
    //
    vector<vector<int> > Faces;
    Faces.resize(numFaces);
    int i,j;
    for (i=0; i<numFaces; i++)
        Faces[i].resize(3);

    // Load data into Faces matrix.
    //
    FILE *fp;
    if ( ( fp = fopen(argv[1],"r") ) == NULL ) {
        fprintf(stderr,"Error opening %s.\n",argv[1]);
        exit(1);
    }
    int lineCount = 0;
    int val1, val2, val3;
    while ( fscanf(fp,"%d %d %d",&val1,&val2,&val3)!=EOF ) {
        Faces[lineCount][0] = val1;
        Faces[lineCount][1] = val2;
        Faces[lineCount][2] = val3;
        lineCount++;
    }
    fclose(fp);
//    printf("lineCount = %d\n",lineCount);

    // Find min and max in Faces matrix.
    // The max "should" be the number of vertices.
    //
    int min = Faces[0][0];
    int max = Faces[0][0];
    for (i=0; i<numFaces; i++) {
        for (j=0; j<3; j++) {
            if ( min > Faces[i][j] )
                min = Faces[i][j];
            if ( max < Faces[i][j] )
                max = Faces[i][j];
        }
    }
//    int numVerts = max; // Use this instead for 1-offset indices
    int numVerts = max + 1;
    printf("Number of vertices = %d\n",numVerts);

    // Make a tally of the number of occurrences of each vertex.
    //
    vector<int> vertexCount(numVerts,0);
    int theVal;
    for (i=0; i<numFaces; i++) {
        for (j=0; j<3; j++) {
            theVal = Faces[i][j];
            vertexCount[theVal]++;
//            vertexCount[theVal-]++; // Use this instead for 1-offset indices
        }
    }

    // Check to make sure that each vertex appears at least three times in the Faces matrix.
    //
    for (i=0; i<numVerts; i++) {
        if ( vertexCount[i] < 3 ) {
            fprintf(stderr,"vertexCount[%d] = %d!\n",i,vertexCount[i]);
            fprintf(stderr,"Each vertex should appear at least three times in the Faces matrix!\n",i,vertexCount[i]);
            exit(2);
        }
    }

    // Make a tally of all occurrences of all edges in the Faces matrix.
    // An edge is defined simply by a pair of vertex indices.
    // Important: the pair of vertex indices are ordered -- the smaller one goes first.
    // Edges[i][j] will be the edge formed by the vertices 'i' and 'j', where i < j.
    //
    vector<list<int> > Edges(numVerts);
    for (i=0; i<numFaces; i++) {
        AddEdge(Edges,Faces[i][0],Faces[i][1]);
        AddEdge(Edges,Faces[i][0],Faces[i][2]);
        AddEdge(Edges,Faces[i][1],Faces[i][2]);
    }

    // Each edge should appear EXACTLY TWICE.
    // If an edge doesn't appear EXACTLY TWICE, alert user.
    // Loop over vertices.
    //
    list<int>::iterator it;
    vector<int> occurrenceCount(numVerts);
    bool foundBadEdge = false;
    int numEdges = 0;
    for (i=0; i<numVerts; i++) {
        // Initialize occurrenceCount to zeros.
        //
        for (j=0; j<numVerts; j++)
            occurrenceCount[j] = 0;

        // Now step through the edges associated with vertex 'i',
        // and count the number of occurrences of each edge.
        //
        for ( it = (Edges[i]).begin(); it != (Edges[i]).end(); ++it )
            occurrenceCount[*it]++;

	// Make sure that each edge appears EXACTLY TWICE.
        //
        for (j=0; j<numVerts; j++) {
            if ( ( occurrenceCount[j] != 0 ) && ( occurrenceCount[j] != 2 ) ) {
                printf("Edges[%d][%d] occurs %d times\n!",i,j,occurrenceCount[j]);
                foundBadEdge = true;
            }
            if ( occurrenceCount[j] != 0 ) {
                numEdges++;
//                printf("Edges[%d][%d] occurs %d times\n!",i,j,occurrenceCount[j]);
            }
        }
    }
    printf("Number of edges    = %d\n\n",numEdges);

    if ( ! foundBadEdge ) {
        printf("No holes detected.\n");
    }

    exit(0);
}
