#include "pinpolyhedron.h"
#include "stopwatch.h"
#include <float.h>         // For FLT_MAX

#define MAX_LINE_LEN 1024

#define SUBTRACT_MIN 0

int CountLinesInFile(char *filename)
{
    FILE *fIN=NULL;
    char dummy[MAX_LINE_LEN];
    int count=0;
    
    if ((fIN=fopen(filename,"r")) == 0) {
        fprintf(stderr,"CountLinesInFile: can't open %s\n",filename);
        exit(1);
    }
    while (fgets(dummy,MAX_LINE_LEN,fIN)!=NULL) count++;

    fclose(fIN);
    return count;
}   
    
int CountWordsInFile(char *filename)
{
    FILE *fIN;
    char dummy[MAX_LINE_LEN];
    
    if ((fIN=fopen(filename,"r")) == 0) {
        fprintf(stderr,"CountWordsInFile: can't open %s\n",filename);
        return -1;
    }
    int count=1;
    if (fscanf(fIN,"%s",dummy) == EOF) {
        fprintf(stderr,"CountWordsInFile: %s is empty!\n",filename);
        return -1;
    }
    while(fscanf(fIN,"%s",dummy)!=EOF)
        count++;
    
    fclose(fIN);
    return count;
}    

PointInPolyhedron *ptpoly=0;

int main(int argc, char *argv[]){
	double (*vert)[3];
	int numvert,numtri,(*tris)[3], i;

    //-------------------------------------------------------------------
    // Check command line arguments.
    
    if ( argc !=4 ) {
         fprintf(stderr,"Need three plain-text files: XYZ coordinates of vertices,\n");
         fprintf(stderr,"vertex indices defining faces, and XYZ coordinates of queries\n");
         exit(1);
    }

    //-------------------------------------------------------------------
    // Read in verts from first file.
    
	numvert = CountWordsInFile(argv[1]) / 3;
//    printf( "Polyhedron Vertices:\n" );
//    printf( "  i:   x   y   z\n");
	vert = ( double (*)[3]) new double [3*numvert];
	float tmpX, tmpY, tmpZ;
	float minX = FLT_MAX, minY=FLT_MAX, minZ=FLT_MAX;

    FILE *fVert;
    if ( ( fVert = fopen(argv[1],"r") ) == NULL ) {
         fprintf(stderr,"Unable to open %s!\n",argv[1]);
         exit(2);     
    }
    for ( i = 0; i < numvert; i++ ) {
        fscanf(fVert,"%g %g %g", &tmpX, &tmpY, &tmpZ );
        vert[i][0] = tmpX;
        vert[i][1] = tmpY;
        vert[i][2] = tmpZ;
//        printf( "%3d:%4g%4g%4g\n", i, vert[i][0], vert[i][1], vert[i][2] );
#if SUBTRACT_MIN
        if ( minX > tmpX ) minX = tmpX;
        if ( minY > tmpY ) minY = tmpY;
        if ( minZ > tmpZ ) minZ = tmpZ;
#endif
    }
    fclose(fVert);
    printf("%d vertices read in\n",numvert);
#if SUBTRACT_MIN
    printf("minX = %g\n",minX);
    printf("minY = %g\n",minY);
    printf("minZ = %g\n",minZ);
    for ( i = 0; i < numvert; i++ ) {
        vert[i][0] -= minX;
        vert[i][1] -= minY;
        vert[i][2] -= minZ;
    }
#endif
//    putchar('\n');

    //-------------------------------------------------------------------
	// Read in trianglar faces from second file.
	
	numtri = CountWordsInFile(argv[2]) / 3;
    tris= (int (*)[3]) new int [3*numtri];
    int min1=INT_MAX, min2=INT_MAX, min3=INT_MAX;
    int minIndex=INT_MAX;
//    printf("About to read in %3d faces...\n",numtri);fflush(NULL);
    FILE *fFace;
    if ( ( fFace = fopen(argv[2],"r") ) == NULL ) {
         fprintf(stderr,"Unable to open %s!\n",argv[2]);
         exit(2);     
    }    
    for(int i=0; i<numtri; i++) {
        fscanf(fFace,"%d %d %d", &tris[i][0], &tris[i][1], &tris[i][2] );
//        printf( "%3d:%3d%4d%4d\n", i, tris[i][0], tris[i][1], tris[i][2] );

        // Decrement data by 1 if it is 1-offset.
        // In the original code, indexing starts from 0, not 1.
        // CONDITIONAL DECREMENTING OF VERTEX INDICES IS PERFORMED BELOW.
        // tris[i][0]--; tris[i][1]--; tris[i][2]--;

        // Keep track of minimum index.
        if ( minIndex > tris[i][0] ) minIndex = tris[i][0];
        if ( minIndex > tris[i][1] ) minIndex = tris[i][1];
        if ( minIndex > tris[i][2] ) minIndex = tris[i][2];
#if 0
        if (min1>tris[i][0]) min1=tris[i][0];
        if (min2>tris[i][1]) min2=tris[i][1];
        if (min3>tris[i][2]) min3=tris[i][2];
#endif
    }
    fclose(fFace);
    printf("%d faces read in\n",numtri);
#if 0
    printf("min1 = %d\n",min1);
    printf("min2 = %d\n",min2);
    printf("min3 = %d\n",min3);
#endif

    //-------------------------------------------------------------------
    // CONDITIONAL DECREMENTING OF VERTEX INDICES.
    // If vertex indexing in the triangular faces appears to be 1-offset,
    // decrement the indices so that they're 0 offset.

    switch ( minIndex ) {
        case 0:
             break;
        case 1: {
             for(int i=0; i<numtri; i++) {
                 tris[i][0]--;
                 tris[i][1]--;
                 tris[i][2]--;
             }
             break;
         }
         default: {
             fprintf(stderr,
                 "Indices in %s appear to be neither 0- nor 1-offset!\n",argv[2]);
             exit(3);
         }
    }

    //-------------------------------------------------------------------
	// Construct a  PointInPolyhedron object.

//printf("About to instantiate a PointInPolyhedron object...\n");fflush(NULL);
    try {
 	    ptpoly=new PointInPolyhedron(vert,numvert,tris,numtri);
    }
    catch(...) {
        fprintf(stderr,"Polyhedron has a topological defect!\n");
        exit(4);
    }
//printf("Finished instantiating a PointInPolyhedron object.\n");fflush(NULL);
	delete [] vert; // Release memory!!
	delete [] tris; // Release memory!!

    //-------------------------------------------------------------------
	// Read in test points 'p' from third file.
	// Test whether each point 'p' is in polyhedron, rt= -1,out; 0,on; 1, in.

    double p[3]={0,0,0};
    stopwatch watch;
    FILE *fQuery;
    if ( ( fQuery = fopen(argv[3],"r") ) == NULL ) {
         fprintf(stderr,"Unable to open %s!\n",argv[3]);
         exit(2);     
    }
    printf("\nRESULTS OF CONTAINMENT TESTS:\n");
    while( fscanf(fQuery,"%g %g %g", &tmpX, &tmpY, &tmpZ ) != EOF ) {
        p[0] = tmpX;
        p[1] = tmpY;
        p[2] = tmpZ;
        
#if SUBTRACT_MIN
        p[0] -= minX;
        p[1] -= minY;
        p[2] -= minZ;
#endif

//        printf( "\n----------->q = %g %g %g\n", p[0], p[1], p[2] );
//        printf( "In = %d\n", ptpoly->isPinPolyhedron (p) );
        printf("%d\n", ptpoly->isPinPolyhedron(p) );
    }
    printf("\n");
    fclose(fQuery);

	delete ptpoly; // Release memory!!
}
